"""Provides the O&M Enviroment class; a subclass of simpy.Environment."""
from __future__ import annotations

import io
import csv
import math
import logging
import datetime as dt
from typing import TYPE_CHECKING
from pathlib import Path
from datetime import datetime, timedelta

import numpy as np
import simpy
import pandas as pd
import pyarrow as pa
import pyarrow.csv  # pylint: disable=W0611
from simpy.events import Event
from pandas.core.indexes.datetimes import DatetimeIndex

import wombat  # pylint: disable=W0611
from wombat.utilities import hours_until_future_hour
from wombat.core.data_classes import parse_date


if TYPE_CHECKING:
    from wombat.windfarm import Windfarm


EVENTS_COLUMNS = [
    "datetime",
    "env_datetime",
    "env_time",
    "agent",
    "action",
    "reason",
    "additional",
    "system_id",
    "system_name",
    "part_id",
    "part_name",
    "system_operating_level",
    "part_operating_level",
    "duration",
    "distance_km",
    "request_id",
    "location",
    "materials_cost",
    "hourly_labor_cost",
    "salary_labor_cost",
    "total_labor_cost",
    "equipment_cost",
    "total_cost",
]


class WombatEnvironment(simpy.Environment):
    """The primary mechanism for powering an O&M simulation. This object has insight
    into all other simulation objects, and controls the timing, date/time stamps, and
    weather conditions.

    Parameters
    ----------
    data_dir : pathlib.Path | str
        Directory where the inputs are stored and where to save outputs.
    weather_file : str
        Name of the weather file. Should be contained within ``data_dir``/weather/, with
        columns "datetime", "windspeed", and, optionally, "waveheight". The datetime
        column should adhere to the following format: "MM/DD/YY HH:MM", in 24-hour time.
    workday_start : int
        Starting time for the repair crew, in 24 hour local time. This can be overridden
        by an ``ServiceEquipmentData`` object that operates outside of the "typical"
        working hours.
    workday_end : int
        Ending time for the repair crew, in 24 hour local time. This can be overridden
        by an ``ServiceEquipmentData`` object that operates outside of the "typical"
        working hours.
    simulation_name : str | None, optional
        Name of the simulation; will be used for naming the log file, by default None.
        If ``None``, then the current time will be used. Will always save to
        ``data_dir``/outputs/logs/``simulation_name``.log.

        .. note: spaces (" ") will be replaced with underscores ("_"), for example:
            "my example analysis" becomes "my_example_analysis".

    start_year : int | None, optional
        Custom starting year for the weather profile, by default None. If ``None`` or
        less than the first year of the weather profile, this will be ignored.
    end_year : int | None, optional
        Custom ending year for the weather profile, by default None. If ``None`` or
        greater than the last year of the weather profile, this will be ignored.
    port_distance : int | float
        The simulation-wide daily travel distance for servicing equipment. This
        should be used as a base setting when multiple or all servicing equipment
        will be operating out of the same base location, but can be individually
        modified.
    non_operational_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level,
        an undefined or later starting date will be overridden for all servicing
        equipment and any modeled port, by default None.
    non_operational_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of prohibited operations. When defined at the environment level,
        an undefined or earlier ending date will be overridden for all servicing
        equipment and any modeled port, by default None.
    reduced_speed_start : str | datetime.datetime | None
        The starting month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level,
        an undefined or later starting date will be overridden for all servicing
        equipment and any modeled port, by default None.
    reduced_speed_end : str | datetime.datetime | None
        The ending month and day, e.g., MM/DD, M/D, MM-DD, etc. for an annualized
        period of reduced speed operations. When defined at the environment level,
        an undefined or earlier ending date will be overridden for all servicing
        equipment and any modeled port, by default None.
    reduced_speed : float
        The maximum operating speed during the annualized reduced speed operations.
        When defined at the environment level, an undefined or faster value will be
        overridden for all servicing equipment and any modeled port, by default 0.0.

    Raises
    ------
    FileNotFoundError
        Raised if ``data_dir`` cannot be found.
    """

    def __init__(
        self,
        data_dir: Path | str,
        weather_file: str,
        workday_start: int,
        workday_end: int,
        simulation_name: str | None = None,
        start_year: int | None = None,
        end_year: int | None = None,
        port_distance: int | float | None = None,
        non_operational_start: str | dt.datetime | None = None,
        non_operational_end: str | dt.datetime | None = None,
        reduced_speed_start: str | dt.datetime | None = None,
        reduced_speed_end: str | dt.datetime | None = None,
        reduced_speed: float = 0.0,
    ) -> None:
        """Initialization."""
        super().__init__()
        self.data_dir = Path(data_dir).resolve()
        if not self.data_dir.is_dir():
            raise FileNotFoundError(f"{self.data_dir} does not exist")

        self.workday_start = int(workday_start)
        self.workday_end = int(workday_end)
        if not 0 <= self.workday_start <= 24:
            raise ValueError("workday_start must be a valid 24hr time before midnight.")
        if not 0 <= self.workday_end <= 24:
            raise ValueError("workday_end must be a valid 24hr time.")
        if self.workday_end <= self.workday_start:
            raise ValueError(
                "Work shifts must end after they start ({self.workday_start}hrs)."
            )

        self.port_distance = port_distance
        self.weather = self._weather_setup(weather_file, start_year, end_year)
        self.weather_dates = self.weather.index.to_pydatetime()
        self.max_run_time = self.weather.shape[0]
        self.shift_length = self.workday_end - self.workday_start

        # Set the environmental consideration parameters
        self.non_operational_start = parse_date(non_operational_start)
        self.non_operational_end = parse_date(non_operational_end)
        self.reduced_speed_start = parse_date(reduced_speed_start)
        self.reduced_speed_end = parse_date(reduced_speed_end)
        self.reduced_speed = reduced_speed

        self.simulation_name = simulation_name
        self._logging_setup()
        self.process(self._log_actions())

    def _register_windfarm(self, windfarm: Windfarm) -> None:
        """Adds the simulation windfarm to the class attributes."""
        self.windfarm = windfarm

    def run(self, until: int | float | Event | None = None):
        """Extends the ``simpy.Environment.run`` method to change the default behavior
        if no argument is passed to ``until``, which will now run a simulation until the
        end of the weather profile is reached.

        Parameters
        ----------
        until : Optional[Union[int, float, Event]], optional
            When to stop the simulation, by default None. See documentation on
            ``simpy.Environment.run`` for more details.
        """
        # If running a paused simulation, then reopen the file and append, but only if
        # the simulation time is lower than the upper bound
        time_check = self.now < self.max_run_time
        if self._events_csv.closed and time_check:  # type: ignore
            self._events_csv = open(self.events_log_fname, "a")
            self._events_writer = csv.DictWriter(
                self._events_csv, delimiter="|", fieldnames=EVENTS_COLUMNS
            )
        if hasattr(self, "windfarm") and self._operations_csv.closed and time_check:
            self._operations_csv: io.TextIOWrapper = open(
                self.operations_log_fname, "a"
            )
            self.windfarm._setup_logger(initial=False)

        if until is None:
            until = self.max_run_time
        elif until > self.max_run_time:
            until = self.max_run_time
        try:
            super().run(until=until)
        except BaseException as e:
            # Flush the logs to so the simulation up to the point of failure is logged
            self._events_writer.writerows(self._events_buffer)
            self._events_buffer.clear()
            self._events_csv.close()
            self._operations_writer.writerows(self._operations_buffer)
            self._operations_buffer.clear()
            self._operations_csv.close()
            raise e

        # Ensure all logged events make it to their target file
        self._events_writer.writerows(self._events_buffer)
        self._events_buffer.clear()
        self._events_csv.close()
        self._operations_writer.writerows(self._operations_buffer)
        self._operations_buffer.clear()
        self._operations_csv.close()

    def _logging_setup(self) -> None:
        """Completes the setup for logging data."""
        if self.simulation_name is None:
            self.simulation_name = simulation = "wombat"
        else:
            simulation = self.simulation_name.replace(" ", "_")
        dt_stamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        events_log_fname = f"{dt_stamp}_{simulation}_events.csv"
        operations_log_fname = f"{dt_stamp}_{simulation}_operations.csv"
        power_potential_fname = f"{dt_stamp}_{simulation}_power_potential.csv"
        power_production_fname = f"{dt_stamp}_{simulation}_power_production.csv"
        metrics_input_fname = f"{dt_stamp}_{simulation}_metrics_inputs.yaml"

        log_path = self.data_dir / "results"
        if not log_path.exists():
            log_path.mkdir()
        self.events_log_fname = log_path / events_log_fname
        self.operations_log_fname = log_path / operations_log_fname
        self.power_potential_fname = log_path / power_potential_fname
        self.power_production_fname = log_path / power_production_fname
        self.metrics_input_fname = log_path / metrics_input_fname

        _dir = self.data_dir / "results"
        if not _dir.is_dir():
            _dir.mkdir()

        self._events_csv = open(self.events_log_fname, "w")
        self._operations_csv = open(self.operations_log_fname, "w")
        self._events_writer = csv.DictWriter(
            self._events_csv, delimiter="|", fieldnames=EVENTS_COLUMNS
        )
        self._events_writer.writeheader()

        self._events_buffer: list[dict] = []
        self._operations_buffer: list[dict] = []

    @property
    def simulation_time(self) -> datetime:
        """Current time within the simulation ("datetime" column within weather)."""
        now = self.now
        minutes = now % 1 * 60
        if now == self.max_run_time:
            _dt = self.weather_dates[math.floor(now - 1)]
            _dt + timedelta(hours=1)
        else:
            _dt = self.weather_dates[math.floor(now)]

        minutes, seconds = math.floor(minutes), math.ceil(minutes % 1 * 60)
        return _dt + timedelta(minutes=minutes, seconds=seconds)

    def is_workshift(self, workday_start: int = -1, workday_end: int = -1) -> bool:
        """Check if the current simulation time is within the windfarm's working hours.

        Parameters
        ----------
        workday_start : int
            A valid hour in 24 hour time, by default -1. This should only be provided
            from an ``ServiceEquipmentData`` object. ``workday_end`` must also be
            provided in order to be used.
        workday_end : int
            A valid hour in 24 hour time, by default -1. This should only be provided
            from an ``ServiceEquipmentData`` object. ``workday_start`` must also be
            provided in order to be used.

        Returns
        -------
        bool
            True if it's valid working hours, False otherwise.
        """
        if -1 in (workday_start, workday_end):
            # Return True if the shift is around the clock
            if self.workday_start == 0 and self.workday_end == 24:
                return True
            return self.workday_start <= self.simulation_time.hour < self.workday_end

        # Return true if the shift is around the clock
        if workday_start == 0 and workday_end == 24:
            return True

        return workday_start <= self.simulation_time.hour < workday_end

    def hour_in_shift(
        self, hour: int, workday_start: int = -1, workday_end: int = -1
    ) -> bool:
        """Checks whether an ``hour`` is within the working hours.

        Parameters
        ----------
        hour : int
            Hour of the day.
        workday_start : int
            A valid hour in 24 hour time, by default -1. This should only be provided
            from an ``ServiceEquipmentData`` object. ``workday_end`` must also be
            provided in order to be used.
        workday_end : int
            A valid hour in 24 hour time, by default -1. This should only be provided
            from an ``ServiceEquipmentData`` object. ``workday_start`` must also be
            provided in order to be used.

        Returns
        -------
        bool
            True if ``hour`` is during working hours, False otherwise.
        """
        if -1 in (workday_start, workday_end):
            return self.workday_start <= hour < self.workday_end
        return workday_start <= hour < workday_end

    def hours_to_next_shift(self, workday_start: int = -1) -> float:
        """Time until the next work shift starts, in hours.

        Parameters
        ----------
        workday_start : int
            A valid hour in 24 hour time, by default -1. This should only be provided
            from an ``ServiceEquipmentData`` object.

        Returns
        -------
        float
            Hours until the next shift starts.
        """
        current = self.simulation_time
        start = self.workday_start if workday_start == -1 else workday_start
        if current.hour < start:
            # difference between now and workday start
            return hours_until_future_hour(current, start)
        elif current.hour == start == 0:
            # Need to manually move forward one whole day to avoid an infinite loop
            return hours_until_future_hour(current, 24)
        else:
            # time to midnight + hour of workday start
            return start + hours_until_future_hour(current, 0)

    @property
    def current_time(self) -> str:
        """Timestamp for the current time as a datetime.datetime.strftime."""
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")

    def date_ix(self, date: dt.datetime | dt.date) -> int:
        """The first index of a future date. This corresponds to the number of hours
        until this dates from the very beginning of the simulation.

        Parameters
        ----------
        date : datetime.datetime | datetime.date
            A date within the environment's simulation range.

        Returns
        -------
        int
            Index of the weather profile corresponds to the first hour of ``date``.
        """
        ix = self.weather.index.get_loc(date.strftime("%Y-%m-%d"))

        # If the index is consecutive a slice is returned, else a numpy array
        if isinstance(ix, slice):
            return ix.start
        return ix[0]

    def _weather_setup(
        self,
        weather_file: str,
        start_year: int | None = None,
        end_year: int | None = None,
    ) -> pd.DataFrame:
        """Reads the weather data from the "<inputs>/weather" directory, and creates the
        ``start_date`` and ``end_date`` time stamps for the simulation.

        This also fills any missing data with zeros and interpolates the values of any
        missing datetime entries.

        Parameters
        ----------
        weather_file : str
            Name of the weather file to be used by the environment. Should be contained
            within ``data_dir/weather``.
        start_year : Optional[int], optional
            Custom starting year for the weather profile, by default None. If ``None``
            or less than the first year of the weather profile, this will be ignored.
        end_year : Optional[int], optional
            Custom ending year for the weather profile, by default None. If ``None`` or
            greater than the last year of the weather profile, this will be ignored.

        Returns
        -------
        pd.DataFrame
            The wind (and  wave) timeseries.
        """
        REQUIRED = ["windspeed", "waveheight"]

        # PyArrow datetime conversion setup
        convert_options = pa.csv.ConvertOptions(
            timestamp_parsers=[
                "%m/%d/%y %H:%M",
                "%m/%d/%y %I:%M",
                "%m/%d/%Y %H:%M",
                "%m/%d/%Y %I:%M",
            ]
        )
        weather = (
            pa.csv.read_csv(
                self.data_dir / "weather" / weather_file,
                convert_options=convert_options,
            )
            .to_pandas()
            .set_index("datetime")
        )
        weather = weather.fillna(0.0)
        weather = weather.resample("H").interpolate(limit_direction="both", limit=5)

        missing = set(REQUIRED).difference(weather.columns)
        if missing:
            raise KeyError(
                "The weather data are missing the following required columns:"
                f" {missing}"
            )

        # Add in the hour of day column for efficient handling within the simulation
        weather = weather.assign(hour=weather.index.hour.astype(float))

        # Create the start and end points
        self.start_datetime = weather.index[0].to_pydatetime()
        self.end_datetime = weather.index[-1].to_pydatetime()
        self.start_year = self.start_datetime.year
        self.end_year = self.end_datetime.year

        if start_year is None and end_year is None:
            return weather

        if start_year is None:
            pass
        elif start_year > self.end_year:
            raise ValueError(
                f"'start_year' ({start_year}) occurs after the last available year"
                f" in the weather data (range: {self.end_year})"
            )
        else:
            # Filter for the provided, validated starting year and update the attribute
            weather = weather.loc[weather.index.year >= start_year]
            self.start_datetime = weather.index[0].to_pydatetime()
            start_year = self.start_year = self.start_datetime.year

        if end_year is None:
            pass
        elif start_year is None and end_year < self.start_year:
            raise ValueError(
                f"The provided 'end_year' ({end_year}) is before the start_year"
                f" ({self.start_year})"
            )
        elif start_year is not None:
            if end_year < start_year:
                raise ValueError(
                    f"The provided 'end_year' ({end_year}) is before the start_year"
                    f" ({start_year})"
                )
            else:
                # Filter for the provided, validated ending year and update
                weather = weather.loc[weather.index.year <= end_year]
                self.end_datetime = weather.index[-1].to_pydatetime()
                self.end_year = self.end_datetime.year
        else:
            # Filter for the provided, validated ending year and update the attribute
            weather = weather.loc[weather.index.year <= end_year]
            self.end_datetime = weather.index[-1].to_pydatetime()
            self.end_year = self.end_datetime.year

        column_order = weather.columns.tolist()
        column_order.insert(0, column_order.pop(column_order.index("hour")))
        column_order.insert(0, column_order.pop(column_order.index("waveheight")))
        column_order.insert(0, column_order.pop(column_order.index("windspeed")))

        return weather.loc[:, column_order]

    @property
    def weather_now(self) -> tuple[float, float, int]:
        """The current weather.

        Returns
        -------
        Tuple[float, float, int]
            Wind, wave, and hour data for the current time.
        """
        # Rounds down because we won't arrive at the next weather event until that hour
        now = int(self.now)
        return self.weather.iloc[now].values

    def weather_forecast(
        self, hours: int | float
    ) -> tuple[DatetimeIndex, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the wind and wave data for the next ``hours`` hours, starting from
        the current hour's weather.

        Parameters
        ----------
        hours : Union[int, float]
            Number of hours to look ahead, rounds up to the nearest hour.

        Returns
        -------
        Tuple[DatetimeIndex, np.ndarray, np.ndarray, np.ndarray]
            The pandas DatetimeIndex, windspeed array, and waveheight array for the
            hours requested, each with shape (``hours`` + 1).
        """
        start = math.floor(self.now)

        # If it's not on the hour, ensure we're looking ``hours`` hours into the future
        end = start + math.ceil(hours) + math.ceil(self.now % 1)

        wind, wave, hour, *_ = self.weather.values[start:end].T
        ix = self.weather.index[start:end]
        return ix, hour, wind, wave

    def log_action(
        self,
        *,
        agent: str,
        action: str,
        reason: str,
        additional: str = "",
        system_id: str = "",
        system_name: str = "",
        part_id: str = "",
        part_name: str = "",
        system_ol: float | int = 0,
        part_ol: float | int = 0,
        duration: float = 0,
        distance_km: float = 0,
        request_id: str = "na",
        location: str = "na",
        materials_cost: int | float = 0,
        hourly_labor_cost: int | float = 0,
        salary_labor_cost: int | float = 0,
        equipment_cost: int | float = 0,
    ) -> None:
        """Formats the logging messages into the expected format for logging.

        Parameters
        ----------
        agent : str
            Agent performing the action.
        action : str
            Action that was taken.
        reason : str
            Reason an action was taken.
        additional : str
            Any additional information that needs to be logged.
        system_id : str
            Turbine ID, ``System.id``, by default "".
        system_name : str
            Turbine name, ``System.name``, by default "".
        part_id : str
            Subassembly, component, or cable ID, ``_.id``, by default "".
        part_name : str
            Subassembly, component, or cable name, ``_.name``, by default "".
        system_ol : float | int
            Turbine operating level, ``System.operating_level``. Use an empty string
            for n/a, by default 0.
        part_ol : float | int
            Subassembly, component, or cable operating level, ``_.operating_level``. Use
            an empty string for n/a, by default 0.
        request_id : str
            The ``RepairManager`` assigned request_id found in
            ``RepairRequest.request_id``, by default "na".
        location : str
            The location of where the event ocurred: should be one of site, port,
            enroute, or system, by default "na".
        duration : float
            Length of time the action lasted, by default 0.
        distance : float
            Distance traveled, in km, if applicable, by default 0.
        materials_cost : Union[int, float], optional
            Total cost of materials for action, in USD, by default 0.
        hourly_labor_cost : Union[int, float], optional
            Total cost of hourly labor for action, in USD, by default 0.
        salary_labor_cost : Union[int, float], optional
            Total cost of salaried labor for action, in USD, by default 0.
        equipment_cost : Union[int, float], optional
            Total cost of equipment for action, in USD, by default 0.
        """
        valid_locations = ("site", "system", "port", "enroute", "na")
        if location not in valid_locations:
            raise ValueError(
                f"Event logging `location` must be one of: {valid_locations}"
            )
        total_labor_cost = hourly_labor_cost + salary_labor_cost
        total_cost = total_labor_cost + equipment_cost + materials_cost
        row = {
            "datetime": dt.datetime.now(),
            "env_datetime": self.simulation_time,
            "env_time": self.now,
            "system_id": system_id,
            "system_name": system_name,
            "part_id": part_id,
            "part_name": part_name,
            "system_operating_level": system_ol,
            "part_operating_level": part_ol,
            "agent": agent,
            "action": action,
            "reason": reason,
            "additional": additional,
            "duration": duration,
            "distance_km": distance_km,
            "request_id": request_id,
            "location": location,
            "materials_cost": materials_cost,
            "hourly_labor_cost": hourly_labor_cost,
            "salary_labor_cost": salary_labor_cost,
            "equipment_cost": equipment_cost,
            "total_labor_cost": total_labor_cost,
            "total_cost": total_cost,
        }
        self._events_buffer.append(row)

    def _log_actions(self):
        """Writes the action log items every 8000 hours."""
        HOURS = 8000
        while True:
            yield self.timeout(HOURS)
            self._events_writer.writerows(self._events_buffer)
            self._events_buffer.clear()

    def load_events_log_dataframe(self) -> pd.DataFrame:
        """Imports the logging file created in ``run`` and returns it as a formatted
        ``pandas.DataFrame``.

        Returns
        -------
        pd.DataFrame
            The formatted logging data from a simulation.
        """
        convert_options = pa.csv.ConvertOptions(
            timestamp_parsers=["%Y-%m-%d %H:%M:%S.%f", "%Y-%m-%d %H:%M:%S"]
        )
        parse_options = pa.csv.ParseOptions(delimiter="|")
        log_df = pa.csv.read_csv(
            self.events_log_fname,
            convert_options=convert_options,
            parse_options=parse_options,
        ).to_pandas()
        if not pd.api.types.is_datetime64_any_dtype(log_df.datetime):
            log_df.datetime = pd.to_datetime(
                log_df.datetime, yearfirst=True, format="mixed"
            )
        if not pd.api.types.is_datetime64_any_dtype(log_df.env_datetime):
            log_df.env_datetime = pd.to_datetime(
                log_df.env_datetime, yearfirst=True, format="mixed"
            )
        log_df = log_df.set_index("datetime").sort_values("datetime")

        return log_df

    def load_operations_log_dataframe(self) -> pd.DataFrame:
        """Imports the logging file created in ``run`` and returns it as a formatted
        ``pandas.DataFrame``.

        Returns
        -------
        pd.DataFrame
            The formatted logging data from a simulation.
        """
        convert_options = pa.csv.ConvertOptions(
            timestamp_parsers=["%Y-%m-%d %H:%M:%S.%f", "%Y-%m-%d %H:%M:%S"]
        )
        parse_options = pa.csv.ParseOptions(delimiter="|")
        log_df = pa.csv.read_csv(
            self.operations_log_fname,
            convert_options=convert_options,
            parse_options=parse_options,
        ).to_pandas()
        if not pd.api.types.is_datetime64_any_dtype(log_df.datetime):
            log_df.datetime = pd.to_datetime(
                log_df.datetime, yearfirst=True, format="mixed"
            )
        if not pd.api.types.is_datetime64_any_dtype(log_df.env_datetime):
            log_df.env_datetime = pd.to_datetime(
                log_df.env_datetime, yearfirst=True, format="mixed"
            )
        log_df = log_df.set_index("datetime").sort_values("datetime")

        return log_df

    def power_production_potential_to_csv(  # type: ignore
        self,
        windfarm: wombat.windfarm.Windfarm,  # type: ignore
        operations: pd.DataFrame | None = None,
        return_df: bool = True,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Creates the power production ``DataFrame`` and optionally returns it.

        Parameters
        ----------
        windfarm : wombat.windfarm.Windfarm
            The simulation's windfarm object.
        operations : Optional[pd.DataFrame], optional
            The operations log ``DataFrame`` if readily available, by default None. If
            ``None``, then it will be created through
            ``load_operations_log_dataframe()``.
        return_df : bool, optional
            Indicator to return the power production for further usage, by default True.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            The power potential and production timeseries data.
        """
        write_options = pa.csv.WriteOptions(delimiter="|")

        if operations is None:
            operations = self.load_operations_log_dataframe().sort_values("env_time")

        turbines = windfarm.turbine_id
        windspeed = self.weather.windspeed
        windspeed = windspeed.loc[operations.env_datetime].values
        potential_df = pd.DataFrame(
            [],
            index=operations.env_datetime,
            columns=["env_time", "env_datetime", "windspeed", "windfarm"]
            + turbines.tolist(),
        )
        potential_df[turbines] = np.vstack(
            [windfarm.system(t_id).power(windspeed) for t_id in turbines]
        ).T
        potential_df = potential_df.assign(
            windspeed=windspeed,
            windfarm=potential_df[turbines].sum(axis=1),
            env_time=operations.env_time.values,
            env_datetime=operations.env_datetime.values,
        )
        pa.csv.write_csv(
            pa.Table.from_pandas(potential_df),
            self.power_potential_fname,
            write_options=write_options,
        )

        # TODO: The actual windfarm production needs to be clipped at each subgraph to
        # the max of the substation's operating capacity and then summed.
        production_df = potential_df.copy()
        production_df[turbines] *= operations[turbines].values
        production_df.windfarm = production_df[turbines].sum(axis=1)
        pa.csv.write_csv(
            pa.Table.from_pandas(production_df),
            self.power_production_fname,
            write_options=write_options,
        )
        if return_df:
            return potential_df, production_df

    def cleanup_log_files(self) -> None:
        """Convenience method to clear the output log files in case a large
        batch of simulations is being run and there are space limitations.

        ... warning:: This shuts down the loggers, so no more logging will be able
            to be performed.
        """
        # NOTE: Everything is wrapped in a try/except clause to protect against failure
        # when inevitably a file has already been deleted on accident, or if in the
        # dataframe generation step, the original logs were deleted

        logging.shutdown()
        if not self._events_csv.closed:
            self._events_csv.close()
        if not self._operations_csv.closed:
            self._operations_csv.close()

        try:
            self.events_log_fname.unlink()
        except FileNotFoundError:
            pass

        try:
            self.operations_log_fname.unlink()
        except FileNotFoundError:
            pass

        try:
            self.power_potential_fname.unlink()
        except FileNotFoundError:
            pass

        try:
            self.power_production_fname.unlink()
        except FileNotFoundError:
            pass

        try:
            self.metrics_input_fname.unlink()
        except FileNotFoundError:
            pass



"""A common location for base class mixins that provide functionality shared across
different types of classes.
"""

from __future__ import annotations

import datetime
from typing import Generator
from functools import partial

import numpy as np
import pandas as pd
from simpy.events import Process, Timeout

from wombat.utilities import HOURS_IN_DAY, check_working_hours
from wombat.utilities.time import calculate_cost
from wombat.core.environment import WombatEnvironment

from .data_classes import (
    Failure,
    PortConfig,
    Maintenance,
    ScheduledServiceEquipmentData,
    UnscheduledServiceEquipmentData,
)


class RepairsMixin:
    """A Mixin class that provides some common repair and timing methods that are used
    across different repair-providing classes, such as: servicing equipment and ports.
    """

    env: WombatEnvironment
    settings: PortConfig | ScheduledServiceEquipmentData | UnscheduledServiceEquipmentData  # noqa: disable=E501

    def initialize_cost_calculators(self, which: str) -> None:
        """Creates the cost calculators for each of the subclasses that will need to
        calculate hourly costs.

        Parameters
        ----------
        which : str
            One of "port" or "equipment" to to indicate how to access equipment costs
        """
        self.calculate_salary_cost = partial(
            calculate_cost,
            rate=self.settings.crew.day_rate,
            n_rate=self.settings.crew.n_day_rate,
            daily_rate=True,
        )
        self.calculate_hourly_cost = partial(
            calculate_cost,
            rate=self.settings.crew.hourly_rate,
            n_rate=self.settings.crew.n_hourly_rate,
        )
        self.calculate_equipment_cost = partial(
            calculate_cost,
            rate=0 if which == "port" else self.settings.equipment_rate,  # type: ignore
            daily_rate=True,
        )

    def _check_working_hours(self, which: str) -> None:
        """Checks the working hours of the port and overrides a default (-1) to
        the ``env`` settings, otherwise hours remain the same.

        Parameters
        ----------
        which : str
            One of "env" or "port" to determine from which overarching environment
            variable should be used to override unset settings.
        """
        if which == "env":
            start = self.env.workday_start
            end = self.env.workday_end
        elif which == "port":
            start = self.port.settings.workday_start  # type: ignore
            end = self.port.settings.workday_end  # type: ignore
        else:
            raise ValueError(
                "Can only set the workday settings from a 'port' or 'env'."
            )

        self.settings._set_environment_shift(
            *check_working_hours(
                start, end, self.settings.workday_start, self.settings.workday_end
            )
        )

    def _is_workshift(self, hour_ix: np.ndarray | float | int) -> np.ndarray | bool:
        """Determines which timestamps are in the servicing equipment's working hours.

        Parameters
        ----------
        hour_ix : np.ndarray | float | int
            The hour of day component of the datetime stamp.

        Returns
        -------
        np.ndarray | bool
            A boolean array for which values in working hours (True), and which values
            are outside working hours (False).
        """
        is_workshift = self.settings.workday_start <= hour_ix
        is_workshift &= hour_ix <= self.settings.workday_end
        return is_workshift

    def wait_until_next_shift(self, **kwargs) -> Generator[Timeout, None, None]:
        """Delays the process until the start of the next shift.

        Yields
        ------
        Generator[Timeout, None, None]
            Delay until the start of the next shift.
        """
        kwargs["additional"] = kwargs.get(
            "additional", "work shift has ended; waiting for next shift to start"
        )
        kwargs["action"] = kwargs.get("action", "delay")

        # Assume at port, unless otherwise indicated
        if "location" not in kwargs:
            kwargs["location"] = "port"

        delay = self.env.hours_to_next_shift(workday_start=self.settings.workday_start)
        salary_cost = self.calculate_salary_cost(delay)
        hourly_cost = self.calculate_hourly_cost(0)
        equpipment_cost = self.calculate_equipment_cost(delay)
        self.env.log_action(
            duration=delay,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equpipment_cost,
            **kwargs,
        )
        yield self.env.timeout(delay)

    def process_repair(
        self, hours: int | float, request_details: Maintenance | Failure, **kwargs
    ) -> Generator[Timeout | Process, None, None]:
        """The logging and timeout process for performing a repair or doing maintenance.

        Parameters
        ----------
        hours : int | float
            The lenght, in hours, of the repair or maintenance task.
        request_details : Maintenance | Failure
            The deatils of the request, this is only being checked for the type.
        kwargs : dict
            Additional parameters to be passed to ``WombatEnvironment.log_action``.

        Yields
        ------
        Generator[Timeout | Process, None, None]
            A ``Timeout`` is yielded of length ``hours``.
        """
        action = "repair" if isinstance(request_details, Failure) else "maintenance"
        salary_cost = self.calculate_salary_cost(hours)
        hourly_cost = self.calculate_hourly_cost(hours)
        equipment_cost = self.calculate_equipment_cost(hours)
        self.env.log_action(
            action=action,
            duration=hours,
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equipment_cost,
            additional=action,
            location="port" if isinstance(self.settings, PortConfig) else "system",
            **kwargs,
        )
        yield self.env.timeout(hours)

    def hours_to_next_operational_date(
        self,
        start_search_date: datetime.datetime | datetime.date | pd.Timestamp,
        exclusion_days: int = 0,
    ) -> float:
        """Calculates the number of hours until the first date that is not a part of
        the ``non_operational_dates`` given a starting datetime and for search criteria.
        Optionally, ``exclusion_days`` can be used to account for a mobilization period
        so that mobilization can occur during the non operational period.

        Parameters
        ----------
        start_search_date : datetime.datetime | datetime.date | pd.Timestamp
            The first date to be considered in the search.
        exclusion_days : int, optional
            The number of days to subtract from the next available datetime that
            represents a non operational action that can occur during the non
            operational period, such as mobilization, by default 0.

        Returns
        -------
        float
            The total number of hours until the next operational date.
        """
        current = self.env.simulation_time

        if isinstance(start_search_date, datetime.date):
            start_search_date = datetime.datetime.combine(
                start_search_date, datetime.datetime.max.time()
            )

        # Get the forecast and filter out dates prior to the start of the search
        all_dates, *_ = self.env.weather_forecast(
            8760
        )  # get the dates for the next year
        dates = all_dates[np.where(all_dates > start_search_date)[0]]
        if dates.size == 0:
            dates = all_dates[-1:]

        # Find the next operational date, and if no available dates, return the time
        # until the end of the forecast period
        date_diff = set(dates.date).difference(self.settings.non_operational_dates_set)
        if not date_diff:
            diff = ((max(dates) - current).days + 1) * HOURS_IN_DAY
            return diff
        next_available = min(date_diff)
        next_available = dates[np.where(dates.date == next_available)[0][0]]

        # Compute the difference between the current time and the future date
        diff = next_available - current
        hours_to_next = diff.days * HOURS_IN_DAY + diff.seconds / 60 / 60
        hours_to_next -= exclusion_days * HOURS_IN_DAY
        hours_to_next = max(0, hours_to_next)
        return hours_to_next



"""General logging methods."""

from __future__ import annotations

import logging
import datetime
from typing import Any
from pathlib import Path
from logging.handlers import MemoryHandler


def setup_logger(
    logger_name: str, log_file: Path, level: Any = logging.INFO, capacity: int = 500
) -> None:
    """Creates the logging infrastructure for a given logging category.

    TODO: Figure out how to type check ``logging.INFO``; ``Callable``?

    Parameters
    ----------
    logger_name : str
        Name to assign to the logger.
    log_file : Path
        File name and path for where the log data should be saved.
    level : Any, optional
        Logging level, by default logging.INFO.
    """
    logger = logging.getLogger(logger_name)
    formatter = logging.Formatter("%(asctime)s|%(name)s|%(levelname)s|%(message)s")
    fileHandler = logging.FileHandler(log_file, mode="w")
    fileHandler.setFormatter(formatter)

    memory_handler = MemoryHandler(capacity=capacity, target=fileHandler)
    memory_handler.setFormatter(formatter)

    logger.setLevel(level)
    # logger.addHandler(fileHandler)
    logger.addHandler(memory_handler)


def format_events_log_message(
    simulation_time: datetime.datetime,
    env_time: float,
    system_id: str,
    system_name: str,
    part_id: str,
    part_name: str,
    system_ol: float | str,
    part_ol: float | str,
    agent: str,
    action: str,
    reason: str,
    additional: str,
    duration: float,
    request_id: str,
    location: str = "na",
    materials_cost: int | float = 0,
    hourly_labor_cost: int | float = 0,
    salary_labor_cost: int | float = 0,
    equipment_cost: int | float = 0,
) -> str:
    """Formats the logging messages into the expected format for logging.

    Parameters
    ----------
    simulation_time : datetime64
        Timestamp within the simulation time.
    env_time : float
        Environment simulation time (``Environment.now``).
    system_id : str
        Turbine ID, ``System.id``.
    system_name : str
        Turbine name, ``System.name``.
    part_id : str
        Subassembly, component, or cable ID, ``_.id``.
    part_name : str
        Subassembly, component, or cable name, ``_.name``.
    system_ol : int | float
        System operating level, ``System.operating_level``. Use an empty string for n/a.
    part_ol : int | float
        Subassembly, component, or cable operating level, ``_.operating_level``. Use an
        empty string for n/a.
    agent : str
        Agent performin the action.
    action : str
        Action that was taken.
    reason : str
        Reason an action was taken.
    additional : str
        Any additional information that needs to be logged.
    duration : float
        Length of time the action lasted.
    request_id : str
        The ``RepairRequest.request_id`` or "na".
    location : str
        The location of where the event ocurred: should be one of site, port,
        enroute, or system, by default "na".
    materials_cost : int | float, optional
        Total cost of materials for action, in USD, by default 0.
    hourly_labor_cost : int | float, optional
        Total cost of hourly labor for action, in USD, by default 0.
    salary_labor_cost : int | float, optional
        Total cost of salaried labor for action, in USD, by default 0.
    equipment_cost : int | float, optional
        Total cost of equipment for action, in USD, by default 0.

    Returns
    -------
    str
        Formatted message for consistent logging.[summary]
    """
    total_labor_cost = hourly_labor_cost + salary_labor_cost
    total_cost = total_labor_cost + equipment_cost + materials_cost
    message = (
        f"{simulation_time}|{env_time:f}|{system_id}|{system_name}|{part_id}"
        f"|{part_name}|{system_ol:f}|{part_ol:f}|{agent}|{action}"
        f"|{reason}|{additional}|{duration:f}|{request_id}|{location}"
        f"|{materials_cost:f}|{hourly_labor_cost:f}|{salary_labor_cost:f}"
        f"|{equipment_cost:f}|{total_labor_cost:f}|{total_cost:f}"
    )
    return message



"""Creates the Turbine class."""
from __future__ import annotations

import logging
from typing import Callable
from operator import mul
from functools import reduce

import numpy as np
import pandas as pd

from wombat.core import RepairManager, WombatEnvironment
from wombat.utilities import IEC_power_curve
from wombat.windfarm.system import Subassembly
from wombat.utilities.utilities import create_variable_from_string


class System:
    """Can either be a turbine or substation, but is meant to be something that consists
    of 'Subassembly' pieces.

    See `here <https://www.sciencedirect.com/science/article/pii/S1364032117308985>`_
    for more information.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        repair_manager: RepairManager,
        t_id: str,
        name: str,
        subassemblies: dict,
        system: str,
    ):
        """Initializes an individual windfarm asset.

        Parameters
        ----------
        env : WombatEnvironment
            The simulation environment.
        repair_manager : RepairManager
            The simulation repair and maintenance task manager.
        t_id : str
            The unique identifier for the asset.
        name : str
            The long form name/descriptor for the system/asset.
        subassemblies : dict
            The dictionary of subassemblies required for the system/asset.
        system : str
            The identifier should be one of "turbine" or "substation" to indicate the
            type of system this will be.

        Raises
        ------
        ValueError
            [description]
        """
        self.env = env
        self.repair_manager = repair_manager
        self.id = t_id
        self.name = name
        self.capacity = subassemblies["capacity_kw"]
        self.subassemblies: list[Subassembly] = []
        self.servicing = self.env.event()
        self.cable_failure = self.env.event()

        # Ensure servicing statuses starts as processed and inactive
        self.servicing.succeed()
        self.cable_failure.succeed()

        system = system.lower().strip()
        self._calculate_system_value(subassemblies)
        if system not in ("turbine", "substation"):
            raise ValueError("'system' must be one of 'turbine' or 'substation'!")

        self._create_subassemblies(subassemblies, system)

    def _calculate_system_value(self, subassemblies: dict) -> None:
        """Calculates the turbine's value based its capex_kw and capacity.

        Parameters
        ----------
        system : str
            One of "turbine" or "substation".
        subassemblies : dict
            Dictionary of subassemblies.
        """
        self.value = subassemblies["capacity_kw"] * subassemblies["capex_kw"]

    def _create_subassemblies(self, subassembly_data: dict, system: str) -> None:
        """Creates each subassembly as a separate attribute and also a list for quick
        access.

        Parameters
        ----------
        subassembly_data : dict
            Dictionary providing the maintenance and failure definitions for at least
            one subassembly named
        system : str
            One of "turbine" or "substation" to indicate if the power curves should also
            be created, or not.
        """
        # Set the subassembly data variables from the remainder of the keys in the
        # system configuration file/dictionary
        exclude_keys = ["capacity_kw", "capex_kw", "power_curve"]
        for key, data in subassembly_data.items():
            if key in exclude_keys:
                continue
            name = create_variable_from_string(key)
            subassembly = Subassembly(self, self.env, name, data)
            setattr(self, name, subassembly)
            self.subassemblies.append(getattr(self, name))

        if self.subassemblies == []:
            raise ValueError(
                "At least one subassembly definition requred for ",
                f"ID: {self.id}, Name: {self.name}.",
            )

        self.env.log_action(
            agent=self.name,
            action=f"subassemblies created: {[s.id for s in self.subassemblies]}",
            reason="windfarm initialization",
            system_id=self.id,
            system_name=self.name,
            system_ol=self.operating_level,
            part_ol=1,
            additional="initialization",
        )

        # If the system is a turbine, create the power curve, if available
        if system == "turbine":
            self._initialize_power_curve(subassembly_data.get("power_curve", None))

    def _initialize_power_curve(self, power_curve_dict: dict | None) -> None:
        """Creates the power curve function based on the ``power_curve`` input in the
        ``subassembly_data`` dictionary. If there is no valid input, then 0 will always
        be reutrned.

        Parameters
        ----------
        power_curve_dict : dict
            The turbine definition dictionary.
        """
        self.power_curve: Callable
        if power_curve_dict is None:
            self.power_curve = IEC_power_curve(pd.Series([0]), pd.Series([0]))
        else:
            power_curve = power_curve_dict["file"]
            try:
                power_curve_file = self.env.data_dir / "turbines" / power_curve
                power_curve = pd.read_csv(power_curve_file)
            except FileNotFoundError:
                power_curve_file = self.env.data_dir / "windfarm" / power_curve
                power_curve = pd.read_csv(power_curve_file)
                logging.warning(
                    "DeprecationWarning: In v0.8, all power curve files must be located"
                    " in: '<library>/turbines"
                )
            power_curve = power_curve.loc[power_curve.power_kw != 0].reset_index(
                drop=True
            )
            bin_width = power_curve_dict.get("bin_width", 0.5)
            self.power_curve = IEC_power_curve(
                power_curve.windspeed_ms,
                power_curve.power_kw,
                windspeed_start=power_curve.windspeed_ms.min(),
                windspeed_end=power_curve.windspeed_ms.max(),
                bin_width=bin_width,
            )

    def interrupt_all_subassembly_processes(
        self, origin: Subassembly | None = None
    ) -> None:
        """Interrupts the running processes in all of the system's subassemblies.

        Parameters
        ----------
        origin : Subassembly
            The subassembly that triggered the request, if the method call is coming
            from a subassembly shutdown event.
        """
        [
            subassembly.interrupt_processes(origin=origin)  # type: ignore
            for subassembly in self.subassemblies
        ]

    @property
    def operating_level(self) -> float:
        """The turbine's operating level, based on subassembly and cable performance.

        Returns
        -------
        float
            Operating level of the turbine.
        """
        if self.cable_failure.triggered and self.servicing.triggered:
            return reduce(mul, [sub.operating_level for sub in self.subassemblies])
        return 0.0

    @property
    def operating_level_wo_servicing(self) -> float:
        """The turbine's operating level, based on subassembly and cable performance,
        without accounting for servicing status.

        Returns
        -------
        float
            Operating level of the turbine.
        """
        if self.cable_failure.triggered:
            return reduce(mul, [sub.operating_level for sub in self.subassemblies])
        return 0.0

    def power(self, windspeed: list[float] | np.ndarray) -> np.ndarray:
        """Generates the power output for an iterable of windspeed values.

        Parameters
        ----------
        windspeed : list[float] | np.ndarrays
            Windspeed values, in m/s.

        Returns
        -------
        np.ndarray
            Power production, in kW.
        """
        return self.power_curve(windspeed)



"""Test the ServiceEquipment class."""
from __future__ import annotations

import numpy as np
import pytest
import numpy.testing as npt

from wombat.windfarm import Windfarm
from wombat.core.library import load_yaml
from wombat.core.data_classes import (
    VALID_EQUIPMENT,
    Maintenance,
    EquipmentMap,
    RepairRequest,
)
from wombat.core.repair_management import RepairManager
from wombat.core.service_equipment import ServiceEquipment, consecutive_groups


def get_items_by_description(
    manager: RepairManager, description: str
) -> list[RepairRequest]:
    """Get all items in the repair manager by description."""
    return [el for el in manager.items if el.details.description == description]


def test_consecutive_groups():
    """Test the consecutive_groups function in a similar way that this would be used."""
    all_clear = np.array(
        [True, True, False, True, True, True, False, True, True, True, True, False]
    )
    windows = consecutive_groups(np.where(all_clear)[0])
    correct_windows = [np.array([0, 1]), np.array([3, 4, 5]), np.array([7, 8, 9, 10])]
    assert len(windows) == len(correct_windows)
    for window, correct_window in zip(windows, correct_windows):
        npt.assert_equal(window, correct_window)

    # Flip the weather window and test that we only have 1 hour openings
    all_clear = ~all_clear
    windows = consecutive_groups(np.where(all_clear)[0])
    correct_windows = [np.array([2]), np.array([6]), np.array([11])]
    assert len(windows) == len(correct_windows)
    for window, correct_window in zip(windows, correct_windows):
        npt.assert_equal(window, correct_window)


def test_service_equipment_init(env_setup):
    """Test the ServiceEquipment initialization."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)

    # TEST 1
    # Check that a basic CTV initialization looks good

    ctv = ServiceEquipment(env, windfarm, manager, "ctv_quick_load.yaml")
    ctv.finish_setup_with_environment_variables()
    ctv_dict = load_yaml(env.data_dir / "vessels", "ctv_quick_load.yaml")

    # Check the basic attribute assignments
    assert ctv.env == env
    assert ctv.windfarm == windfarm
    assert ctv.manager == manager
    assert not ctv.onsite
    assert not ctv.enroute
    assert not ctv.at_system
    assert not ctv.at_port
    assert not ctv.at_site
    assert ctv.current_system is None

    # Check that the starting and ending year is the same as the input file
    assert ctv.settings.start_year == ctv_dict["start_year"]
    assert ctv.settings.end_year == ctv_dict["end_year"]

    # Check that the default inputs are set to the environment settings
    assert ctv.settings.workday_start == env.workday_start
    assert ctv.settings.workday_end == env.workday_end

    # TEST 2
    # Check that a CTV with invalid start years gets corrected
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_invalid_years.yaml")
    ctv.finish_setup_with_environment_variables()

    # Check the basic attribute assignments
    assert ctv.env == env
    assert ctv.windfarm == windfarm
    assert ctv.manager == manager
    assert not ctv.onsite
    assert not ctv.enroute
    assert not ctv.at_system
    assert not ctv.at_site
    assert not ctv.at_port
    assert ctv.current_system is None

    # Check that the starting and ending year match the environment limits
    assert ctv.settings.start_year == env.start_year
    assert ctv.settings.end_year == env.end_year

    # Check that the default inputs are set to the environment settings
    assert ctv.settings.workday_start == env.workday_start
    assert ctv.settings.workday_end == env.workday_end

    # TEST 3
    # Check the initializtion for an unscheduled vessel for a downtime-basis
    hlv = ServiceEquipment(env, windfarm, manager, "hlv_downtime.yaml")
    hlv.finish_setup_with_environment_variables()
    hlv_dict = load_yaml(env.data_dir / "vessels", "hlv_downtime.yaml")

    # Check the basic attribute assignments
    assert hlv.env == env
    assert hlv.windfarm == windfarm
    assert hlv.manager == manager
    assert not hlv.onsite
    assert not hlv.enroute
    assert not hlv.at_system
    assert not hlv.at_site
    assert not hlv.at_port
    assert hlv.current_system is None

    # Check that the hours are the custom set hours
    assert hlv.settings.workday_start == hlv_dict["workday_start"]
    assert hlv.settings.workday_end == hlv_dict["workday_end"]

    # Check that the vessel is registered correctly in the repair manager
    assert manager.downtime_based_equipment.is_running
    assert not manager.request_based_equipment.is_running  # No vessels to run or check
    assert manager.downtime_based_equipment.LCN == [
        EquipmentMap(hlv_dict["strategy_threshold"], hlv)
    ]
    for capability in VALID_EQUIPMENT:
        if capability == "LCN":
            assert len(getattr(manager.downtime_based_equipment, capability)) == 1
        else:
            assert getattr(manager.downtime_based_equipment, capability) == []

    # TEST 4
    # Check the initializtion for an unscheduled vessel for a request-basis
    hlv = ServiceEquipment(env, windfarm, manager, "hlv_requests.yaml")
    hlv.finish_setup_with_environment_variables()
    hlv_dict = load_yaml(env.data_dir / "vessels", "hlv_requests.yaml")

    # Check the basic attribute assignments
    assert hlv.env == env
    assert hlv.windfarm == windfarm
    assert hlv.manager == manager
    assert not hlv.onsite
    assert not hlv.enroute
    assert not hlv.at_system
    assert not hlv.at_site
    assert not hlv.at_port
    assert hlv.current_system is None

    # Check that the hours are the custom set hours
    assert hlv.settings.workday_start == hlv_dict["workday_start"]
    assert hlv.settings.workday_end == hlv_dict["workday_end"]

    # Check that the vessel is registered correctly in the repair manager
    assert (
        manager.downtime_based_equipment.is_running
    )  # Same manager, just another HLV added with a different strategy
    assert manager.request_based_equipment.is_running
    assert manager.request_based_equipment.LCN == [
        EquipmentMap(hlv_dict["strategy_threshold"], hlv)
    ]
    for capability in ("CTV", "SCN", "CAB", "RMT", "DRN", "DSV"):
        assert getattr(manager.request_based_equipment, capability) == []


def test_port_distance_setup(env_setup):
    """Test that the environment port_distance is passed through correctly."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)

    # Check that a None for WombatEnvironment.port_distance does nothing
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_quick_load.yaml")
    ctv.finish_setup_with_environment_variables()
    assert env.port_distance is None
    assert ctv.settings.port_distance == 0

    # Check that <=0 does not override no inputs to servicing equipment settings
    env.port_distance = -0.1
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_quick_load.yaml")
    ctv.finish_setup_with_environment_variables()
    assert env.port_distance == -0.1
    assert ctv.settings.port_distance == 0

    # Check that a real port_distance overrides the default servicing equipment settings
    env.port_distance = 10
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_quick_load.yaml")
    ctv.finish_setup_with_environment_variables()
    assert env.port_distance == ctv.settings.port_distance == 10

    # Check that an already set value to port_distance will not be overridden
    ctv = ServiceEquipment(
        env, windfarm, manager, "ctv_quick_load_with_port_distance.yaml"
    )
    ctv.finish_setup_with_environment_variables()
    assert ctv.settings.port_distance == 1


def test_calculate_salary_cost(env_setup):
    """Test the ``calculate_salary_cost`` method."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_wages.yaml")
    ctv.finish_setup_with_environment_variables()
    ctv_dict = load_yaml(env.data_dir / "vessels", "ctv_wages.yaml")
    ctv_crew = ctv_dict["crew"]

    # Test a small, even number
    n_hours = 4
    cost = ctv.calculate_salary_cost(n_hours)
    assert cost == n_hours / 24 * ctv_crew["day_rate"] * ctv_crew["n_day_rate"]

    # Test a multi-day number
    n_hours = 74.333
    cost = ctv.calculate_salary_cost(n_hours)
    assert cost == n_hours / 24 * ctv_crew["day_rate"] * ctv_crew["n_day_rate"]


def test_calculate_hourly_cost(env_setup):
    """Test the ``calculate_hourly_cost`` method."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)
    ctv = ServiceEquipment(env, windfarm, manager, "ctv_wages.yaml")
    ctv.finish_setup_with_environment_variables()
    ctv_dict = load_yaml(env.data_dir / "vessels", "ctv_wages.yaml")
    ctv_crew = ctv_dict["crew"]

    n_hours = 4
    cost = ctv.calculate_hourly_cost(n_hours)
    assert cost == n_hours * ctv_crew["hourly_rate"] * ctv_crew["n_hourly_rate"]

    n_hours = 74.333
    cost = ctv.calculate_hourly_cost(n_hours)
    assert cost == n_hours * ctv_crew["hourly_rate"] * ctv_crew["n_hourly_rate"]


def test_calculate_equipment_cost(env_setup):
    """Test the ``calculate_equipment_cost`` method."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)
    ctv = ServiceEquipment(env, windfarm, manager, "ctv.yaml")
    ctv.finish_setup_with_environment_variables()
    ctv_dict = load_yaml(env.data_dir / "vessels", "ctv.yaml")

    n_hours = 4
    cost = ctv.calculate_equipment_cost(n_hours)
    assert cost == n_hours / 24 * ctv_dict["equipment_rate"]

    n_hours = 74.333
    cost = ctv.calculate_equipment_cost(n_hours)
    assert cost == n_hours / 24 * ctv_dict["equipment_rate"]


def test_onsite_scheduled_equipment_logic(env_setup_full_profile):
    """Test the simulation logic of a scheduled CTV."""
    env = env_setup_full_profile
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout_simulation.csv", manager)

    ctv = ServiceEquipment(env, windfarm, manager, "ctv.yaml")
    ctv.finish_setup_with_environment_variables()
    fsv = ServiceEquipment(env, windfarm, manager, "fsv_scheduled.yaml")
    fsv.finish_setup_with_environment_variables()
    hlv = ServiceEquipment(env, windfarm, manager, "hlv_scheduled.yaml")
    hlv.finish_setup_with_environment_variables()

    # Start the simulation to ensure everything is in place as required
    env.run(1)
    assert (
        ctv.at_port
        is ctv.at_site
        is ctv.at_system
        is ctv.transferring_crew
        is ctv.enroute
        is ctv.onsite
        is False
    )
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.enroute
        is hlv.onsite
        is False
    )

    # Fast-forward to the start of the environment workshift to ensure all are
    # non-operational (CTV has no tasks, nor mobilization to check)
    env.run(7)
    assert (
        ctv.at_port
        is ctv.at_site
        is ctv.at_system
        is ctv.transferring_crew
        is ctv.enroute
        is ctv.onsite
        is False
    )
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.enroute
        is hlv.onsite
        is False
    )

    # Check that at Day 15 there are 6 turbine generator maintenance tasks requested
    timeout = 15 * 24 + 1
    env.run(timeout)
    assert len(manager.items) == 6
    assert all(isinstance(el.details, Maintenance) for el in manager.items)
    assert all(el.details.description == "ctv call" for el in manager.items)

    # The first maintenance task is at day 15 for the CTV, fast forward to this point in
    # time to check for operations to be running and inspect every step of the task
    timeout += 6 + 1 / 60  # Shift start is 7AM, but it's already 1AM
    env.run(timeout + 0.01)
    assert ctv.onsite is ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert not ctv.enroute
    assert ctv.current_system == "S00T1"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Move to just before the 15 minute transfer is complete to ensure it hasn't ended
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T1"
    assert not ctv.enroute

    # Move to the end of the 15 minute to ensure the transfer is complete
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T1"
    assert ctv.enroute is ctv.at_port is ctv.transferring_crew is False

    # The weather window is clear on 1/16/2002, so there will not be weather delays, so
    # the CTV should be transferring the crew back onto the vessel
    timeout += 1
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.at_port is ctv.enroute is False
    assert ctv.current_system == "S00T1"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Check the transfer is still progressing
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T1"
    assert ctv.enroute is ctv.at_port is False
    assert len(manager.items) == 5

    ## TURBINE 2 GENERATOR CTV repair

    # Move to the end of the 15 minute to ensure the transfer is complete, and there is
    # no travel time between site objects, so the crew should be transferring again
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.current_system == "S00T2"
    assert ctv.at_port is ctv.enroute is False

    # Move to just before the 15 minute transfer is complete to ensure it hasn't ended
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T2"
    assert ctv.at_port is ctv.enroute is False

    # Move to the end of the 15 minute to ensure the transfer is complete
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T2"
    assert ctv.enroute is ctv.at_port is ctv.transferring_crew is False

    # The weather window is clear on 1/16/2002, so there will not be weather delays, so
    # the CTV should be transferring the crew back onto the vessel
    timeout += 1
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.enroute is ctv.at_port is False
    assert ctv.current_system == "S00T2"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Check the transfer is still progressing
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T2"
    assert ctv.enroute is ctv.at_port is False
    assert len(manager.items) == 4

    ## TURBINE 3 GENERATOR CTV repair

    # Move to the end of the 15 minute to ensure the transfer is complete, and there is
    # no travel time between site objects, so the crew should be transferring again
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.current_system == "S00T3"
    assert ctv.enroute is ctv.at_port is False

    # Move to just before the 15 minute transfer is complete to ensure it hasn't ended
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T3"
    assert ctv.enroute is ctv.at_port is False

    # Move to the end of the 15 minute to ensure the transfer is complete
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T3"
    assert ctv.enroute is ctv.at_port is ctv.transferring_crew is False

    # The weather window is clear on 1/16/2002, so there will not be weather delays, so
    # the CTV should be transferring the crew back onto the vessel
    timeout += 1
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.enroute is ctv.at_port is False
    assert ctv.current_system == "S00T3"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Check the transfer is still progressing
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S00T3"
    assert ctv.enroute is ctv.at_port is False
    assert len(manager.items) == 3

    ## TURBINE 4 GENERATOR CTV repair

    # Move to the end of the 15 minute to ensure the transfer is complete, and there is
    # no travel time between site objects, so the crew should be transferring again
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.current_system == "S01T4"
    assert ctv.enroute is ctv.at_port is False

    # Move to just before the 15 minute transfer is complete to ensure it hasn't ended
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T4"
    assert ctv.enroute is ctv.at_port is False

    # Move to the end of the 15 minute to ensure the transfer is complete
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.onsite is ctv.at_system is True
    assert ctv.current_system == "S01T4"
    assert ctv.enroute is ctv.transferring_crew is False

    # The weather window is clear on 1/16/2002, so there will not be weather delays, so
    # the CTV should be transferring the crew back onto the vessel
    timeout += 1
    env.run(timeout)
    assert ctv.onsite is ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.enroute is ctv.at_port is False
    assert ctv.current_system == "S01T4"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Check the transfer is still progressing
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T4"
    assert ctv.enroute is ctv.at_port is False
    assert len(manager.items) == 2

    ## TURBINE 5 GENERATOR CTV repair

    # Move to the end of the 15 minute to ensure the transfer is complete, and there is
    # no travel time between site objects, so the crew should be transferring again
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.onsite is ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.current_system == "S01T5"
    assert ctv.enroute is ctv.at_port is False

    # Move to just before the 15 minute transfer is complete to ensure it hasn't ended
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T5"
    assert ctv.enroute is ctv.at_port is False

    # Move to the end of the 15 minute to ensure the transfer is complete
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T5"
    assert ctv.enroute is ctv.at_port is ctv.transferring_crew is False

    # The weather window is clear on 1/16/2002, so there will not be weather delays, so
    # the CTV should be transferring the crew back onto the vessel
    timeout += 1
    env.run(timeout)
    assert ctv.onsite is ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.enroute is ctv.at_port is False
    assert ctv.current_system == "S01T5"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Check the transfer is still progressing
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T5"
    assert ctv.enroute is ctv.at_port is False
    assert len(manager.items) == 1

    ## TURBINE 6 GENERATOR CTV repair

    # Move to the end of the 15 minute to ensure the transfer is complete, and there is
    # no travel time between site objects, so the crew should be transferring again
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.onsite is ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.current_system == "S01T6"
    assert ctv.enroute is ctv.at_port is False

    # Move to just before the 15 minute transfer is complete to ensure it hasn't ended
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T6"
    assert ctv.enroute is ctv.at_port is False

    # Move to the end of the 15 minute to ensure the transfer is complete
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T6"
    assert ctv.enroute is ctv.transferring_crew is False

    # The weather window is clear on 1/16/2002, so there will not be weather delays, so
    # the CTV should be transferring the crew back onto the vessel
    timeout += 1
    env.run(timeout)
    assert ctv.onsite is ctv.at_site is ctv.at_system is ctv.transferring_crew is True
    assert ctv.enroute is ctv.at_port is False
    assert ctv.current_system == "S01T6"
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.enroute
        is fsv.transferring_crew
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.onsite
        is hlv.enroute
        is hlv.transferring_crew
        is False
    )

    # Check the transfer is still progressing
    timeout += 13 / 60
    env.run(timeout)
    assert ctv.transferring_crew is ctv.onsite is ctv.at_site is ctv.at_system is True
    assert ctv.current_system == "S01T6"
    assert ctv.enroute is ctv.at_port is False
    assert len(manager.items) == 0

    # The CTV process will repeat every 15 days (as this is a maintenance task), but no
    # more events will be checked as this process is considered "vetted".
    timeout += 2 / 60
    env.run(timeout)
    assert ctv.at_port and ctv.onsite
    assert ctv.current_system is None
    assert (
        ctv.at_site
        is ctv.transferring_crew
        is ctv.at_site
        is ctv.enroute
        is ctv.at_system
        is False
    )

    current = env.simulation_time
    assert current.date().day == 16
    assert current.date().month == 1
    assert current.date().year == 2002
    assert current.hour == 16
    assert current.minute == 1


@pytest.mark.skip(reason="The timing of the failures needs to be updated")
def test_scheduled_equipment_logic(env_setup_full_profile):
    """Test the simulation logic of a scheduled HLV."""
    env = env_setup_full_profile
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout_simulation.csv", manager)

    # ctv = ServiceEquipment(env, windfarm, manager, "ctv.yaml")
    fsv = ServiceEquipment(env, windfarm, manager, "fsv_scheduled.yaml")
    fsv.finish_setup_with_environment_variables()
    hlv = ServiceEquipment(env, windfarm, manager, "hlv_scheduled.yaml")
    hlv.finish_setup_with_environment_variables()

    # Start the simulation to ensure everything is in place as required
    env.run(1)
    # assert (
    #     ctv.at_port is ctv.at_system
    #     is ctv.transferring_crew
    #     is ctv.enroute
    #     is (not ctv.onsite)
    #     is False
    # )
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.enroute
        is hlv.onsite
        is False
    )

    # Moving on to the HLV, scheduled in May with 60 days mobilization, and
    # default working hours
    timeout = env.date_ix(hlv.settings.operating_dates[0]) + 8
    mobilization_timeout = timeout - 24 * 60

    env.run(mobilization_timeout - 1)
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.enroute
        is hlv.onsite
        is False
    )
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )

    env.run(mobilization_timeout + 1)
    assert hlv.enroute
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.onsite
        is False
    )
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )

    env.run(timeout - 1)  # starting hours
    assert hlv.enroute
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.onsite
        is False
    )
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )

    # Check for the correct number of items before the HLV is operating
    N_days = np.floor(env.now / 24)
    correct_N_ctv_items = (
        np.floor(N_days / 15) * 6
    )  # calls every 15 days for 6 turbines
    correct_N_hlv_items = (
        np.floor(N_days / 50) * 6
    )  # calls every 50 days for 6 turbines
    correct_N_fsv_items = (
        np.floor(N_days / 25) * 6
    )  # calls every 25 days for 6 turbines
    correct_N_items = correct_N_ctv_items + correct_N_hlv_items + correct_N_fsv_items
    n_ctv_items = len(get_items_by_description(manager, "ctv call"))
    n_hlv_items = len(get_items_by_description(manager, "hlv call"))
    n_fsv_items = len(get_items_by_description(manager, "fsv call"))
    assert len(
        manager.items == correct_N_items
    )  # all smaller requests should have been
    assert n_ctv_items == correct_N_ctv_items
    assert n_hlv_items == correct_N_hlv_items
    assert n_fsv_items == correct_N_fsv_items

    # HLV repairs are 3 hours long
    # Repair 1 has operating hours 8-16 with inclimate weather the whole first day, so
    # when the HLV attempts to transfer the crew, it will go back to port.
    # TODO: Have a better logic to pre-empt wasted travel like this
    env.run(timeout + 1)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.transferring_crew is hlv.enroute is hlv.at_site is hlv.at_system is False
    assert len(get_items_by_description(manager, "hlv call")) == 11

    # HLV should be in the same spot at the end of the shift
    timeout += 8
    env.run(timeout - 0.1)
    assert hlv.at_port is hlv.at_site is True
    assert hlv.transferring_crew is hlv.enroute is hlv.at_system is hlv.onsite is False
    assert len(get_items_by_description(manager, "hlv call")) == 11

    env.run(timeout + 0.1)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False

    # The next shift has a clear weather window for the first two hours so the repair
    # should be partially complete at that point, but require an additional 15 minutes
    # for both crew transfer and completion after a windy hour
    timeout += 16
    env.run(timeout - 0.1)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert len(get_items_by_description(manager, "hlv call")) == 11

    env.run(timeout + 0.1)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 11

    timeout += 14 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 11

    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is True
    assert hlv.at_port is hlv.enroute is hlv.transferring_crew is False
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 11

    # Move to just past the end of the repair when the crew transfer is ongoing
    timeout += 5  # 3 hour repair + 1 hour delay
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 11

    timeout += 13 / 60
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 11

    # Now move to the 2nd repair for the day on S00T2 (no travel time between turbines)

    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    timeout += 13 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is True
    assert hlv.at_port is hlv.enroute is hlv.transferring_crew is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Move to just past the end of the repair (no delays) to ensure crew is transferring
    timeout += 2
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Move to just after the crew transfer, at which time the HLV will return to port
    timeout += 16 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.enroute is hlv.transferring_crew is hlv.at_system is hlv.at_site is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Now, check the logic for repair # 3 on S00T3

    # Move to just before the start of the shift to ensure it's starting correctly
    timeout += 16 - 3 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.enroute is hlv.transferring_crew is hlv.at_system is hlv.at_site is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Ensure crew is transferring
    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Ensure crew is still transferring
    timeout += 13 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # TODO: This is where the timing breaks due to now mistimed weather and shift delays
    # with new logic

    # Ensure repair is processed with a 2 hour delay
    print(env.now)
    timeout += 3 + 2 + 2 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Ensure crew is transferring to the vessel
    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 10

    # Ensure crew is transferring to the next turbine
    timeout += 14 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Ensure crew stopped transferring
    timeout += 14 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is True
    assert hlv.at_port is hlv.enroute is hlv.transferring_crew is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Ensure repair is hit with a delay at the end of the shift
    timeout += 2
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Check the crew is transferred
    timeout += 1 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    timeout += 13 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.at_port is hlv.enroute is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    timeout += 1 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Wait until the start of the next shift and check for the crew transferring
    timeout += 16
    env.run(timeout)
    assert hlv.at_port is hlv.onsite
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Check the crew is unable to transfer and there is a shift delay until 5/5 at 14:00
    # TODO: The hlv will travel to site as an edge case in the crew_transfer delay
    timeout += 24
    env.run(timeout)
    assert hlv.onsite is hlv.at_site is hlv.at_system is True
    assert hlv.transferring_crew is hlv.enroute is hlv.at_port is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Check we haven't gone anywhere
    timeout += 6 - 1 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.at_site is hlv.onsite is True
    assert hlv.at_port is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Will be transferring crew now
    timeout += 1 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Check that the maintenance is now going for the remained 1 hour after the transfer
    timeout += 1 + 13 / 60
    env.run(timeout)
    assert hlv.at_system is hlv.at_site is hlv.onsite is True
    assert hlv.at_port is hlv.enroute is hlv.transferring_crew is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Will be transferring crew back to the HLV
    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 8

    # Repair 5 on S01T5 will be delayed overnight due to insufficient time to transfer
    # the crew back and forth with a safety buffer

    # Will be waiting overnight at port
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.enroute is hlv.at_site is hlv.at_system is hlv.transferring_crew is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 7

    # Delayed until 8am
    timeout += 16 + 29 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.enroute is hlv.at_system is hlv.at_site is hlv.transferring_crew is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 7

    # Transferring crew to the turbine
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 7

    # After the 3 hour repair and prior crew transfer, crew is being transferred back
    timeout += 3 + 14 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 7

    # Repair 6, S01T6 with no weather or shift delays

    # Crew is transferred to the next turbine
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "hlv call")) == 6

    # After a crew transfer, and 3 hour repair, the crew is being transferred back
    timeout += 3 + 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.onsite is hlv.at_site is hlv.transferring_crew is True
    assert hlv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "hlv call")) == 6

    # Repair 7 will start on the next day due to time constraints for ensuring a safe
    # window for crew transfer, then will complete without delay on the next day

    # Crew is transferred, and the HLV is back at port waiting to start on 5/6 8AM
    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 5

    # Timeout to 5/6 at 8:01AM for crew transferring to the turbine
    timeout += 1 + 16  # 1 hour left in shift + 16 hours overnight
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 5

    # Crew is transferring back to the vessel after a completed repair
    timeout += 3 + 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "hlv call")) == 5

    # Repair 8, S00T2 has no weather or shift delays

    # Crew is transferring back to the vessel after a completed repair
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 4

    # Crew is transferring back to the vessel after a completed repair
    timeout += 3 + 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "hlv call")) == 4

    # Repair 9, S00T3 will start the next day due to the safety buffer and be processed
    # without delays

    # Crew is back at port
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 3

    # Crew is still at port just before the start of the shift
    timeout += 1 + 16 - 2 / 60  # 1 hour remainder + overnight: 5/8 at 7:59AM
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 3

    # Crew is transferring to the turbine
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "hlv call")) == 3

    # Crew is transferring back to the HLV
    timeout += 3 + 2 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "hlv call")) == 3

    # Repair 10, S01T4 will process without delays

    # Crew is transferring to the turbine
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 2

    # Crew is still at the turbine
    timeout += 3
    env.run(timeout)
    assert hlv.at_system is hlv.at_site is hlv.onsite is True
    assert hlv.at_port is hlv.enroute is hlv.transferring_crew is False
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 2

    # Crew is transferring to the HLV
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "hlv call")) == 2

    # Repair 11, S01T5 is delayed to start until the next day because of the safety,
    # but is unable to transfer the crew safely until 5/10 at 13:00 so is effectively
    # stuck at port

    # HLV goes back to port due to the safety buffer
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # HLV is at port after the delay still due to inclimate weather
    timeout += 1 + 16  # 1 hour remainder + 16 hours overnight
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # HLV travels to site, but is unable to transfer crew just yet
    timeout += 24
    env.run(timeout)
    assert hlv.at_system is hlv.at_site is hlv.at_site is hlv.onsite is True
    assert hlv.at_port is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # HLV is transferring crew to the turbine after a 5 hour delay
    timeout += 5
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # HLV is transferring crew back to the HLV after processing 2.75 hours of the repair
    timeout += 2.75
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # The HLV is back at port
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # Check that the Field Support Vessel has been deployed at midnight to arrive for
    # the start of its first workshift
    timeout += 8
    env.run(timeout)
    assert fsv.enroute
    assert (
        fsv.at_system
        is fsv.at_site
        is fsv.onsite
        is fsv.transferring_crew
        is fsv.at_port
        is False
    )
    assert len(get_items_by_description(manager, "fsv call")) == 30

    # The HLV is transferring the crew to the turbine
    timeout += 8
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # The HLV is transferring the crew back to the HLV
    # 15 minutes transfer + 45 minutes to finish the repair
    timeout += (13 + 45) / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "hlv call")) == 1

    # Repair 12 S01T6 runs without delays

    # The HLV is transferring the crew to the turbine
    timeout += 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # The crew is transferring back to the HLV
    timeout += 3 + 15 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.enroute is False
    assert hlv.at_system is hlv.at_site is hlv.onsite is hlv.transferring_crew is True
    assert hlv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # Repair 13 S01T4 will finish without delalys

    # The HLV is back at port because there are no more requests.
    timeout += 2 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # The HLV will have no more repairs for the remainder of its visit, so check the
    # start of each remaining shift

    # Move to 8:01 AM the following day
    timeout += 19 + 30 / 60
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.onsite is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # The HLV will remain at port
    timeout += 24
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # The HLV will remain at port
    timeout += 24
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # The HLV will remain at port
    timeout += 24
    env.run(timeout)
    assert hlv.at_port is hlv.onsite is True
    assert hlv.at_system is hlv.at_site is hlv.transferring_crew is hlv.enroute is False
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # The HLV will no longer be available and is off to another charter
    # It is enroute because it's technically be chartered, but not
    timeout += 24
    env.run(timeout)
    assert (
        hlv.at_system
        is hlv.onsite
        is hlv.transferring_crew
        is hlv.at_site
        is hlv.at_port
        is hlv.enroute
        is False
    )
    assert hlv.current_system is None
    assert len(get_items_by_description(manager, "hlv call")) == 0

    # Field Support Vessel Checks
    # This is going to only check in on the first day and last day of processes
    # due to the extensive timing checks performed above for the HLV, which cover
    # numerous edge cases for repair timing, and so we mostly care about the continuous
    # shift timing checks. In addition because the FSV only has a waveheight
    # restriction, the only weather delays should occur on 6/28 when there are
    # waves >= 1.5m.

    # The FSV be traveling still
    assert fsv.enroute
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.transferring_crew
        is False
    )
    assert fsv.current_system is None

    # This is not 30 due to delays in the timing caused by repairs when the maintenance
    # timeout clock is not ticking
    assert len(get_items_by_description(manager, "fsv call")) == 30

    # The FSV should not have arrived on 5/31 at 23:59
    timeout += 16 * 24 - 8 - 2 / 60
    env.run(timeout)
    assert fsv.enroute
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.onsite
        is fsv.transferring_crew
        is False
    )
    assert fsv.current_system is None
    # More of the requests should have been submitted at this stage
    assert len(get_items_by_description(manager, "fsv call")) == 33

    # The FSV should now be on site for its first request and tranferring the crew
    timeout += 2 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.onsite is fsv.transferring_crew is fsv.at_site is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 33

    # The FSV should be starting the repair after its 30 minute transfer
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 33

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T1"
    # Normally, this number will be decremented, but another request was submitted
    assert len(get_items_by_description(manager, "fsv call")) == 33

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 32

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 32

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 32

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 31

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 31

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 31

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 30

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 30

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 30

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 29

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 29

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 29

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 28

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 28

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 28

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 27

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 27

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 27

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 26

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 26

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 26

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"
    # The final 3 requests will have been submitted, now that they've caught up with
    # their delays
    assert len(get_items_by_description(manager, "fsv call")) == 27

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 27

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 27

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 26

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 26

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 26

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 25

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 25

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 25

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 24

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 24

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 24

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 23

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 23

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T1"
    assert len(get_items_by_description(manager, "fsv call")) == 23

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 22

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 22

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T2"
    assert len(get_items_by_description(manager, "fsv call")) == 22

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 21

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 21

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"
    assert len(get_items_by_description(manager, "fsv call")) == 21

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 20

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 20

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T4"
    assert len(get_items_by_description(manager, "fsv call")) == 20

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 19

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 19

    # Tranferring the crew back to the FSV
    timeout += 2
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T5"
    assert len(get_items_by_description(manager, "fsv call")) == 19

    # Transferring crew to the next turbine
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is fsv.transferring_crew is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 18

    # Maintenance
    timeout += 30 / 60
    env.run(timeout)
    assert fsv.at_system is fsv.at_site is fsv.onsite is True
    assert fsv.at_port is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system == "S01T6"
    assert len(get_items_by_description(manager, "fsv call")) == 18

    # Now that the timing continued overnight, let's ensure the last day looks good too
    timeout += 25 * 24 - (3 + 30 / 60)
    env.run(timeout)
    assert fsv.at_port is fsv.onsite is True
    assert fsv.at_system is fsv.at_site is fsv.enroute is fsv.transferring_crew is False
    assert fsv.current_system is None
    assert len(get_items_by_description(manager, "fsv call")) == 0

    # Now check that the FSV has left
    timeout += 24
    env.run(timeout)
    assert (
        fsv.at_system
        is fsv.at_site
        is fsv.at_port
        is fsv.enroute
        is fsv.transferring_crew
        is fsv.onsite
        is False
    )
    assert fsv.current_system is None
    assert len(get_items_by_description(manager, "fsv call")) == 0


def test_unscheduled_service_equipment_call(env_setup_full_profile):
    """Tests the calling of downtime-based and requests-based service equipment. This
    test will only consider that the equipment is mobilized when required, arrives at
    site, and leaves as the previous tests will cover repair logics.
    """
    env = env_setup_full_profile
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout_simulation.csv", manager)

    # ctv = ServiceEquipment(env, windfarm, manager, "ctv.yaml")
    fsv = ServiceEquipment(env, windfarm, manager, "fsv_requests.yaml")
    fsv.finish_setup_with_environment_variables()
    hlv = ServiceEquipment(env, windfarm, manager, "hlv_downtime.yaml")
    hlv.finish_setup_with_environment_variables()

    # Start the simulation to ensure everything is in place as required
    env.run(1)
    assert (
        fsv.at_port
        is fsv.at_site
        is fsv.at_system
        is fsv.transferring_crew
        is fsv.enroute
        is fsv.onsite
        is False
    )
    assert (
        hlv.at_port
        is hlv.at_site
        is hlv.at_system
        is hlv.transferring_crew
        is hlv.enroute
        is hlv.onsite
        is False
    )

    # The FSV has a 10 request threshold, with a request being submitted simultaneously
    # by each turbine every 25 days, so at day 50, the FSV will be mobilized
    timeout = 50 * 24 + 1 / 60
    env.run(timeout)
    assert fsv.enroute
    assert (
        fsv.transferring_crew
        is fsv.at_system
        is fsv.at_site
        is fsv.onsite
        is fsv.at_port
        is False
    )
    assert fsv.current_system is None

    # After 21 days the FSV will arrive at site, but have to wait until the start of the
    # work shift before repairs can happen
    timeout += 21 * 24
    env.run(timeout)
    assert fsv.at_port is fsv.onsite is True
    assert fsv.transferring_crew is fsv.at_system is fsv.at_site is fsv.enroute is False
    assert fsv.current_system is None

    # Check that the FSV is working at the start of the work shift
    timeout += 9  # 1 hour weather delay
    env.run(timeout)
    assert fsv.onsite is fsv.at_site is fsv.transferring_crew is fsv.at_system is True
    assert fsv.at_port is fsv.enroute is False
    assert fsv.current_system == "S00T3"

    # The FSV should no longer be on site after 28 days, plus the time to finish the
    # repair that it's started
    timeout += 28 * 24 + 2 + 1  # 28 days + 2 hour repair + 1 hour in crew transfer
    env.run(timeout)
    assert (
        fsv.transferring_crew
        is fsv.at_system
        is fsv.onsite
        is fsv.at_site
        is fsv.at_port
        is fsv.enroute
        is False
    )
    assert fsv.current_system is None

    # The first HLV call is at 4118.374184568947 hours when S00T1's generator has a
    # catastrophic failure putting the windfarm at 83.3% operations, which is less than
    # the 90% threshold. However, because the timing will be delayed during repairs,
    # realized timeout will be at 4154.386685 hours
    timeout = 4154.386685
    env.run(timeout + 1)
    assert hlv.enroute
    assert (
        hlv.transferring_crew
        is hlv.at_system
        is hlv.at_site
        is hlv.onsite
        is hlv.at_port
        is False
    )
    assert hlv.current_system is None

    # Test that the HLV was successfully mobilized
    timeout += 60 * 24
    env.run(timeout + 1 / 60)
    assert hlv.transferring_crew is hlv.at_site is hlv.at_system is hlv.onsite is True
    assert hlv.enroute is hlv.at_port is False
    assert hlv.current_system == "S00T1"

    # Ensure it's still here at the end
    timeout += 30 * 24 - 1 / 60
    env.run(timeout)
    assert hlv.onsite is hlv.at_site is True
    assert hlv.transferring_crew is hlv.at_port is hlv.enroute is False
    assert hlv.at_system
    assert hlv.current_system == "S01T6"



"""Initializes the simulation modules, classes, and functions."""


from .data_classes import (
    Maintenance,
    Failure,
    SubassemblyData,
    RepairRequest,
    ServiceEquipmentData,
    ScheduledServiceEquipmentData,
    UnscheduledServiceEquipmentData,
    FixedCosts,
    FromDictMixin,
    EquipmentMap,
    StrategyMap,
    PortConfig,
)
from .mixins import RepairsMixin
from .environment import WombatEnvironment
from .post_processor import Metrics
from .repair_management import RepairManager
from .port import Port
from .service_equipment import ServiceEquipment
from .simulation_api import Configuration, Simulation



"""Provides the Subassembly class."""

from __future__ import annotations

from typing import Generator

import simpy

from wombat.core import (
    Failure,
    Maintenance,
    RepairRequest,
    SubassemblyData,
    WombatEnvironment,
)


class Subassembly:
    """A major system composes the turbine or substation objects."""

    def __init__(
        self,
        system,
        env: WombatEnvironment,
        s_id: str,
        subassembly_data: dict,
    ) -> None:
        """Creates a subassembly object that models various maintenance and failure
        types.

        Parameters
        ----------
        system : wombat.windfarm.System
            The system containing subassembly object(s).
        env : WombatEnvironment
            The simulation environment.
        s_id : str
            A unique identifier for the subassembly within the system.
        subassembly_data : dict
            A dictionary to be passed to ``SubassemblyData`` for creation and
            validation.
        """
        self.env = env
        self.system = system
        self.id = s_id

        subassembly_data = {**subassembly_data, "system_value": self.system.value}
        self.data = SubassemblyData.from_dict(subassembly_data)
        self.name = self.data.name

        self.operating_level = 1.0
        self.broken = self.env.event()
        self.broken.succeed()  # start the event as inactive

        self.processes = dict(self._create_processes())

    def _create_processes(self):
        """Creates the processes for each of the failure and maintenance types.

        Yields
        ------
        Tuple[Union[str, int], simpy.events.Process]
            Creates a dictionary to keep track of the running processes within the
            subassembly.
        """
        for level, failure in self.data.failures.items():
            yield level, self.env.process(self.run_single_failure(failure))

        for i, maintenance in enumerate(self.data.maintenance):
            yield f"m{i}", self.env.process(self.run_single_maintenance(maintenance))

    def recreate_processes(self) -> None:
        """If a turbine is being reset after a tow-to-port repair, then all processes
        are assumed to be reset to 0, and not pick back up where they left off.
        """
        self.processes = dict(self._create_processes())

    def interrupt_processes(self, origin: Subassembly | None = None) -> None:
        """Interrupts all of the running processes within the subassembly except for the
        process associated with failure that triggers the catastrophic failure.

        Parameters
        ----------
        origin : Subassembly
            The subassembly that triggered the request, if the method call is coming
            from a subassembly shutdown event. If provided, and it is the same as the
            current subassembly, then a try/except flow is used to ensure the process
            that initiated the shutdown is not interrupting itself.
        """
        if origin is not None and id(origin) == id(self):
            for _, process in self.processes.items():
                try:
                    process.interrupt()
                except RuntimeError:  # Process initiating process can't be interrupted
                    pass
            return

        for _, process in self.processes.items():
            process.interrupt()

    def interrupt_all_subassembly_processes(self) -> None:
        """Thin wrapper for ``system.interrupt_all_subassembly_processes``."""
        self.system.interrupt_all_subassembly_processes(origin=self)

    def trigger_request(self, action: Maintenance | Failure):
        """Triggers the actual repair or maintenance logic for a failure or maintenance
        event, respectively.

        Parameters
        ----------
        action : Maintenance | Failure
            The maintenance or failure event that triggers a ``RepairRequest``.
        """
        which = "maintenance" if isinstance(action, Maintenance) else "repair"
        self.operating_level *= 1 - action.operation_reduction
        if action.operation_reduction == 1:
            self.broken = self.env.event()
            self.interrupt_all_subassembly_processes()

        # Remove previously submitted requests if a replacement is required
        if action.replacement:
            _ = self.system.repair_manager.purge_subassembly_requests(
                self.system.id, self.id
            )

        # Automatically submit a repair request
        # NOTE: mypy is not caught up with attrs yet :(
        repair_request = RepairRequest(  # type: ignore
            system_id=self.system.id,
            system_name=self.system.name,
            subassembly_id=self.id,
            subassembly_name=self.name,
            severity_level=action.level,
            details=action,
        )
        repair_request = self.system.repair_manager.register_request(repair_request)
        self.env.log_action(
            system_id=self.system.id,
            system_name=self.system.name,
            part_id=self.id,
            part_name=self.name,
            system_ol=self.system.operating_level,
            part_ol=self.operating_level,
            agent=self.name,
            action=f"{which} request",
            reason=action.description,
            additional=f"severity level {action.level}",
            request_id=repair_request.request_id,
        )
        self.system.repair_manager.submit_request(repair_request)

    def run_single_maintenance(self, maintenance: Maintenance) -> Generator:
        """Runs a process to trigger one type of maintenance request throughout the
        simulation.

        Parameters
        ----------
        maintenance : Maintenance
            A maintenance category.

        Yields
        ------
        simpy.events. HOURS_IN_DAY
            Time between maintenance requests.
        """
        while True:
            hours_to_next = maintenance.frequency
            if hours_to_next == 0:
                remainder = self.env.max_run_time - self.env.now
                try:
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            while hours_to_next > 0:
                start = -1  # Ensure an interruption before processing is caught
                try:
                    # Wait until these events are triggered and back to operational
                    yield (
                        self.system.servicing & self.system.cable_failure & self.broken
                    )

                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0
                    self.trigger_request(maintenance)

                except simpy.Interrupt:
                    if not self.broken.triggered:
                        # The subassembly had to restart the maintenance cycle
                        hours_to_next = 0
                    else:
                        # A different process failed, so subtract the elapsed time
                        # only if it had started to be processed
                        hours_to_next -= 0 if start == -1 else self.env.now - start

    def run_single_failure(self, failure: Failure) -> Generator:
        """Runs a process to trigger one type of failure repair request throughout the
        simulation.

        Parameters
        ----------
        failure : Failure
            A failure classification.

        Yields
        ------
        simpy.events. HOURS_IN_DAY
            Time between failure events that need to request a repair.
        """
        while True:
            hours_to_next = failure.hours_to_next_failure()
            if hours_to_next is None:
                remainder = self.env.max_run_time - self.env.now
                try:
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now
                continue
            while hours_to_next > 0:  # type: ignore
                start = -1  # Ensure an interruption before processing is caught
                try:
                    yield (
                        self.system.servicing & self.system.cable_failure & self.broken
                    )
                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0
                    self.trigger_request(failure)

                except simpy.Interrupt:
                    if not self.broken.triggered:
                        # The subassembly had to be replaced so reset the timing
                        hours_to_next = 0
                    else:
                        # A different process failed, so subtract the elapsed time
                        # only if it had started to be processed
                        hours_to_next -= 0 if start == -1 else self.env.now - start
